#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#ifdef TCL_IRSIM
#include <tk.h>
#endif

#ifndef _DEFS_H
#include "defs.h"
#endif
#ifndef _GLOBALS_H
#include "globals.h"
#endif
#ifndef _NET_MACROS_H
#include "net_macros.h"
#endif
#ifndef _SUBCKT_H
#include "subckt.h"
#endif

#ifdef USER_SUBCKT 

typedef struct _Sub
{
    userSubCircuit *subckt;	/* subcircuit (possibly static) */
    int		   inst;	/* Number of instantiations */
#ifndef TCL_IRSIM
    struct _Sub	   *next;	/* list of commands in bucket */
#endif
} SubCircuit;

extern userSubCircuit subs[];
extern tptr rd_tlist;

#ifdef TCL_IRSIM

extern Tcl_Interp *irsiminterp;

/*----------------------------------------------*/
/* Tcl procedure characters representing states	*/
/*----------------------------------------------*/

static pot2ch[] = {'0', 'x', 'z', '1'};

/*------------------------------*/
/* Method 1:			*/
/* Make use of Tcl Hash tables	*/
/*------------------------------*/

Tcl_HashTable substbl;

/* First add any C-compiled user subcircuits */

public void init_subs(userSubCircuit *subckts)
{
    userSubCircuit *s;
    SubCircuit *sl;
    Tcl_HashEntry *he;
    int newptr;

    Tcl_InitHashTable(&substbl, TCL_STRING_KEYS);

    for (s = subckts; s->name != NULL; s++)
    {
	sl = (SubCircuit *)malloc(sizeof(SubCircuit));
	sl->subckt = s;
	sl->inst = 0 ;

	he = Tcl_CreateHashEntry(&substbl, s->name, &newptr);
	Tcl_SetHashValue(he, (ClientData)sl);
    }
}

/* Instantiate a user subcircuit */

public userSubCircuit *subckt_instantiate(char *sname, int *inst)
{
    int newptr;
    SubCircuit *s;
    Tcl_HashEntry *he;

    he = Tcl_CreateHashEntry(&substbl, sname, &newptr);
    s = (SubCircuit *)Tcl_GetHashValue(he);

    if (s != NULL) {
	s->inst++;
	*inst = s->inst;
	return s->subckt;
    }
    return NULL;
}

#else

/*------------------------------*/
/* Method 2:			*/
/* Simple fixed-size hash table */
/*------------------------------*/

#define SUBSTBLSIZE	16

private SubCircuit *substbl[SUBSTBLSIZE];

private int HashSub();

public void init_subs(userSubCircuit *subckts)
{
    register int n;
    register userSubCircuit *s;
    register SubCircuit *sl;

    for (n=0; n < SUBSTBLSIZE; n++)
	substbl[n] = NULL;
    for (s = subckts; s->name != NULL; s++) {
	sl = (SubCircuit *)malloc(sizeof(SubCircuit));
	sl->subcircuit = s;
	sl->inst = 0;
	n = HashSub(s->name);
	sl->next = substbl[n];
	substbl[n] = sl;
    }
}

private int HashSub(char *name)
{
    register int  hashcode = 0;

    do
	hashcode = (hashcode << 1) ^ (*name | 0x20);
    while (*(++name));
    return (((hashcode >= 0) ? hashcode : ~hashcode) % SUBSTBLSIZE);
}

private	userSubCircuit subckt_instantiate(char *sname, int *inst)
{
    register int n;
    register SubCircuit *s;

    n = HashSub(sname);
    
    for (s = substbl[n]; s != NULL; s = s->next) {
	if (str_eql(s->name, sname) == 0) {
	    s->inst++;
	    *inst = s->inst;
	    return s->subcircuit;
	}
    }
    return NULL;
}

#endif	/* !TCL_IRSIM */
	
/*
 *------------------------------------------------------------------
 * Evalutate a subcircuit model
 *------------------------------------------------------------------
 */

public void subckt_model_C(tptr t)
{
    int i;
    SubcktT *subptr = (SubcktT *)(t->subptr);
    short nins = subptr->subckt->ninputs;
    short nouts = subptr->subckt->noutputs;
    nptr *nodes = subptr->nodes;
    uptr udata = subptr->udata;
#ifdef TCL_IRSIM
    int result;
#endif
    
    char *in, *out = NULL;
    double *delay;

    in  = (char *) malloc((nins + 1) * sizeof(char));
    out = (char *) malloc((nouts + 1) * sizeof(char));
    delay = (double *) malloc(nouts * sizeof(double));

    for (i = 0; i < nins; i++)
	in[i] = nodes[i]->npot;
    for (i = 0; i < nouts; i++)
	out[i] = nodes[i+nins]->npot;

    if (subptr->subckt->model != NULL)
	(*subptr->subckt->model)(in, out, delay, udata);
    else
    {

#ifdef TCL_IRSIM
	/* The Tcl version of IRSIM allows subcircuit to be defined as	*/
	/* Tcl procedures.  This procedure must return a list whose	*/
	/* first element is a string of all the output values, followed	*/
	/* by time delays to all of the outputs.			*/

	for (i = 0; i < nins; i++)
	    in[i] = pot2ch[in[i]];
	in[nins] = '\0';
	for (i = 0; i < nouts; i++)
	    out[i] = pot2ch[out[i]];
	out[nouts] = '\0';

	result = Tcl_VarEval(irsiminterp, subptr->subckt->name, " ",
			in, " ", out, NULL);
	if (result == TCL_OK)
	{
	    int listlen;
	    double dval;
	    char *newoutvals;
	    Tcl_Obj *elem, *rlist;

	    rlist = Tcl_GetObjResult(irsiminterp);
	    result = Tcl_ListObjLength(irsiminterp, rlist, &listlen);
	    if (result == TCL_OK && listlen == 1 + nouts)
	    {
		result = Tcl_ListObjIndex(irsiminterp, rlist, 0, &elem);
		newoutvals = Tcl_GetString(elem);
		if (strlen(newoutvals) == nouts)
		{
		    for (i = 0; i < nouts; i++)
		    {
			result = Tcl_ListObjIndex(irsiminterp, rlist, i + 1, &elem);
			result = Tcl_GetDoubleFromObj(irsiminterp, elem, &delay[i]);
			out[i] = newoutvals[i];
		    }

		    /* Clear the interpreter result */
		    Tcl_ResetResult(irsiminterp);
		}
		else nouts = 0;
	    }
	    else nouts = 0;
	}
	else nouts = 0;
    }
#endif

    for (i = 0; i < nouts; i++ ) {
	switch (out[i]) {
	    case LOW:
	    	QueueFVal(nodes[i+nins], HIGH, (double) 1.0, delay[i]);
	    	QueueFVal(nodes[i+nins+nouts], HIGH, (double) 1.0, delay[i]);
		break;
	    case X:
	    	QueueFVal(nodes[i+nins], X, (double) 1.0, delay[i]);
	    	QueueFVal(nodes[i+nins+nouts], X, (double) 1.0, delay[i]);
		break;
	    case HIGH_Z:
	    	QueueFVal(nodes[i+nins], HIGH, (double) 1.0, delay[i]);
	    	QueueFVal(nodes[i+nins+nouts], LOW, (double) 1.0, delay[i]);
		break;
	    case HIGH:
	    	QueueFVal(nodes[i+nins], LOW, (double) 1.0, delay[i]);
	    	QueueFVal(nodes[i+nins+nouts], LOW, (double) 1.0, delay[i]);
		break;
	}
    }

    free(in); 
    free(out); 
    free(delay);
}

/*
 *------------------------------------------------------------------
 * Create a new subcircuit from the .sim file line
 * "x <inputs> <outputs> [<parameters>] <subcircuit_name>"
 *
 * The subcircuit definition is assumed to be already known;
 * otherwise, this routine returns an error result (-1).
 * Return 0 on success.
 *------------------------------------------------------------------
 */

public int newsubckt(targc, targv)
   int  targc;
   char *targv[];
{
    userSubCircuit *subcircuit;
    SubcktT *subptr, *subptr2;
    int i, n, ninputs, noutputs, rt;
    float rtf;
    ufun init;
    register tptr t = 0;
    tptr t2; lptr l;
    char *sname, *out_name_Ub, *out_name_D;
    int subckt_out_counter = 0;
    nptr *nodes;
    lptr diodes = (lptr)NULL, d;
    uptr usrData = NULL;
    int inst, result = 0;

    subcircuit = subckt_instantiate(targv[targc - 1], &inst);
    if (subcircuit == NULL) {
	lprintf(stderr, "\tError: subcircuit \"%s\" is not defined!\n",
			targv[targc - 1]);
	return -1;
    }

    ninputs = subcircuit->ninputs;
    noutputs = subcircuit->noutputs;
    sname = subcircuit->name;

    if (targc < ninputs + noutputs + 2) {
	lprintf(stderr, "\tError: wrong # args %d\n", targc);
	lprintf(stderr, "\tsubcircuit %s has %d input nodes, %d output nodes\n",
		subcircuit->name, ninputs, noutputs);
	return -1;
    } 

    out_name_Ub = (char *)malloc(strlen(targv[targc-1]) + 20);
    out_name_D  = (char *)malloc(strlen(targv[targc-1]) + 20);

    targc--;
    targv++;
    nodes = (nptr *)malloc((ninputs + 2 * noutputs) * sizeof(nptr));
      
    lprintf(stdout, "defining new subcircuit \"%s\" instance %d #i:%d #o:%d\n",
		subcircuit->name, inst, ninputs, noutputs);

    if (subcircuit->init)
       	usrData = (uptr)(*subcircuit->init)();
	    
    for (i = 0; i < ninputs; i++) {	/* create new "transistor" */
	NEW_TRANS(t);		
	NEW_SUBCKT(subptr);
	t->subptr = (char *)subptr;
	t->ttype = SUBCKT; 
	t->gate = RsimGetNode( targv[i] );
	subptr->nodes = nodes; 
	t->source = VDD_node; 
	t->drain = t->gate;

	/* link it to the list */
	t->scache.t = rd_tlist; rd_tlist = t; 
	t->r = requiv(RESIST, (int)2, (int)(1000000 * LAMBDACM));
	subptr->subckt = subcircuit;

	/* All this can be referenced by a pointer to a subckt structure */
	NEW_LINK(d); d->xtor = t; d->next = diodes;
	subptr->ndiode = d; diodes = d;
	nodes[i] = t->gate;
	subptr->udata = usrData;
    } 
    for (l = subptr->ndiode; l != NULL; l = l->next) {
	t2 = l->xtor;
	subptr2 = (SubcktT *)(t2->subptr);
	subptr2->ndiode = diodes;
    }

    for (n = i; i < targc - 1; i++, n++) {

	/* define two transistors for each output : */

	sprintf(out_name_Ub, "%sUb_%d_%d", sname, inst, subckt_out_counter);
	sprintf(out_name_D , "%sD_%d_%d", sname, inst, subckt_out_counter++);
	if (find(out_name_Ub) || find(out_name_D)) {
	    lprintf(stderr,"Error: nodes named %s or %s already exist\n",
			out_name_Ub, out_name_D);
	    result = -1;
	    goto donesubckt;
	}

	NEW_TRANS(t);	/* Add pmos output driver */
	t->ttype    = PCHAN;
	t->gate     = RsimGetNode(out_name_Ub);
	nodes[n]    = t->gate;
	t->drain    = VDD_node;
	t->source   = RsimGetNode(targv[i]);
	t->scache.t = rd_tlist; /* link it to the list */
	rd_tlist    = t;
	rtf 	    = subcircuit->res[i];
	t->r        = requiv(RESIST, (int)2, (int)(rtf * LAMBDACM));

	NEW_TRANS(t);	/* Add nmos output driver */
	t->ttype    = NCHAN;
	t->gate     = RsimGetNode(out_name_D);
	t->drain    = RsimGetNode(targv[i]);
	nodes[n + noutputs]    = t->gate;
	t->source   = GND_node;
	t->scache.t = rd_tlist; /* link it to the list */
	rd_tlist    = t;
	rtf 	    = subcircuit->res[i + noutputs];
	t->r        = requiv(RESIST, (int)2, (int)(rtf * LAMBDACM));
    } /* for loop over outputs */

donesubckt:
    free(out_name_D);
    free(out_name_Ub);
    return result;
}

#endif /* USER_SUBCKT */
